#-------------------------------------------------------------------------
#	CSNDSHELL.SH
#
#  -----------------------------------------------------------------------
#  Copyright 2009  Alfred Steffens Jr.
#  -----------------------------------------------------------------------
#  -----------------------------------------------------------------------
#    Copying Permission:
#
#    Csndshell is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation.
#
#    Csndshell is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Csndshell (see file called "COPYING"); if not, write to
#    the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
#    Boston, MA 02111-1307  USA
#  ------------------------------------------------------------------------
#	Revision History:
#
#   Release on internet as ver. 1.0.1, Sep 24, 2009, A. Steffens Jr.
#   Expect script, Jan. 13, 2008,  A. Steffens Jr.
#	New file, Feb. 18, 2007,	A. Steffens Jr.
#-------------------------------------------------------------------------

CSNDSHELL_VERSION=1.0.1

#
#	first is first -- you must give some arguments
#
cmdlinestring="$@"
alen=${#cmdlinestring}
if [ $alen -eq 0 ]; then

	echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
	echo
	echo '   No arguments given'
	echo '   Usage:'
	echo '      sh <csndshell.sh> ascore=<score file> <node=TARGET> <node=TARGET> ... '
	echo '      or'
	echo '      sh <csndshell.sh> <cnsd_config_file>'
	echo
	echo '   node targets:'
	echo '      node_score=MTX'
	echo
	echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
	exit 1
fi



#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#	Directory Paths
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#
#
#	The default directory for csndshell.sh is in /usr/local/share/csndshell
#	All the auxilliary tools are referenced under this path.
#	Add the following environment variables to your .bashrc (or .profile)
#	if the installer didn't already do it:
#
#		export PATH_TO_CSNDSHELL=<path>
#		export CSND=$PATH_TO_CSNDSHELL/csndshell.sh
#		export RECORD=$PATH_TO_CSNDSHELL/record/record.sh
#
#	If the PATH_TO_CSNDSHELL is not defined, then the following path will
#	be used.  But it is only visible within the script--you wont be able
#	to call csndshell.sh from you interactive shell with it.
#
plen=${#PATH_TO_CSNDSHELL}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_CSNDSHELL=/usr/local/share/csndshell
	# PATH_TO_CSNDSHELL=.
	echo "   *info*   CSNDSHELL path is $PATH_TO_CSNDSHELL"
fi
#
#   Check that this path exists
#
if [ ! -d $PATH_TO_CSNDSHELL ]; then
    echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
    echo
    echo "ERROR: PATH_TO_CSNDSHELL: $PATH_TO_CSNDSHELL not found"
    echo
    echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
    exit 1;
fi

FULL_PATH_TO_CSNDSHELL="$PATH_TO_CSNDSHELL/csndshell.sh"

#
#	Where to find ascore.awk
#
#	export PATH_TO_ASCORE=<directory path>
#
#
plen=${#PATH_TO_ASCORE}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_ASCORE=$PATH_TO_CSNDSHELL/ascore
	echo "   *info*   ASCORE path is $PATH_TO_ASCORE"
fi
FULL_PATH_TO_ASCORE="$PATH_TO_ASCORE/ascore.awk"


#
# 	Where to find cs2sort.awk
#
#	export PATH_TO_CS2SORT=<directory path>
#
#
plen=${#PATH_TO_CS2SORT}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_CS2SORT=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   CS2SORT path is $PATH_TO_CS2SORT"
fi
FULL_PATH_TO_CS2SORT="$PATH_TO_CS2SORT/cs2sort.awk"


#
# 	Where to find pre_cs2mtx.awk
#
#	export PATH_TO_PRE_CS2MTX=<directory path>
#
#
plen=${#PATH_TO_PRE_CS2MTX}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_PRE_CS2MTX=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   PRE_CS2MTX path is $PATH_TO_PRE_CS2MTX"
fi
FULL_PATH_TO_PRE_CS2MTX="$PATH_TO_PRE_CS2MTX/pre_cs2mtx.awk"


#
# 	Where to find cs2mtx.awk
#
#	export PATH_TO_CS2MTX=<directory path>
#
#
plen=${#PATH_TO_CS2MTX}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_CS2MTX=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   CS2MTX path is $PATH_TO_CS2MTX"
fi
FULL_PATH_TO_CS2MTX="$PATH_TO_CS2MTX/cs2mtx.awk"

#
# 	Where to find post_cs2mtx.awk
#
#	export PATH_TO_POST_CS2MTX=<directory path>
#
#
plen=${#PATH_TO_POST_CS2MTX}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_POST_CS2MTX=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   POST_CS2MTX path is $PATH_TO_POST_CS2MTX"
fi
FULL_PATH_TO_POST_CS2MTX="$PATH_TO_POST_CS2MTX/post_cs2mtx.awk"

#
# 	Where to find malign.awk
#
#	export PATH_TO_MALIGN=<directory path>
#
#
plen=${#PATH_TO_MALIGN}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_MALIGN=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   MALIGN path is $PATH_TO_MALIGN"
fi
FULL_PATH_TO_MALIGN="$PATH_TO_MALIGN/malign.awk"

#
# 	Where to find mtxheader.awk
#
#	export PATH_TO_MTXHEADER=<directory path>
#
#
plen=${#PATH_TO_MTXHEADER}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_MTXHEADER=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   MTXHEADER path is $PATH_TO_MTXHEADER"
fi
FULL_PATH_TO_MTXHEADER="$PATH_TO_MTXHEADER/mtxheader.awk"


#
# 	Where to find run2a
#
#	export PATH_TO_MIDI=<directory path>
#
#
plen=${#PATH_TO_MIDI}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_MIDI=$PATH_TO_CSNDSHELL/mscore
	echo "   *info*   MIDI path is $PATH_TO_MIDI"
fi
FULL_PATH_TO_MIDI="$PATH_TO_MIDI/run2a"


#
# 	Where to find the remote PC with an SSH login
#
#	export PATH_TO_RMIDI=<directory path>
#
#
plen=${#PATH_TO_RMIDI}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_RMIDI=apsteffe@spica
	echo "   *info*   RMIDI path is $PATH_TO_RMIDI"
fi
plen=${#SUBPATH_TO_RMIDI}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	SUBPATH_TO_RMIDI=/work/apsteffe
	echo "   *info*   RMIDI sub path is $SUBPATH_TO_RMIDI"
fi
FULL_PATH_TO_RMIDI="$PATH_TO_RMIDI:$SUBPATH_TO_RMIDI"

#
# 	Where to find the record script
#
#	export PATH_TO_RECORD=<directory path>
#
#
plen=${#PATH_TO_RECORD}
if [ $plen -eq 0 ]; then
	#
	# or, change the dot below to the directory path
	#
	PATH_TO_RECORD=$PATH_TO_CSNDSHELL/record
	echo "   *info*   RECORD script path is $PATH_TO_RECORD"
fi
FULL_PATH_TO_RECORD="$PATH_TO_RECORD/record.sh"


FULL_PATH_TO_SNDINFO=/usr/local/bin/sndinfo


#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#  Required Software Tools
#
	if [ ! -f $FULL_PATH_TO_ASCORE ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_ASCORE not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $FULL_PATH_TO_CS2SORT ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_CS2SORT not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $FULL_PATH_TO_PRE_CS2MTX ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_PRE_CS2MTX not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $FULL_PATH_TO_CS2MTX ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_CS2MTX not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $FULL_PATH_TO_MALIGN ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_MALIGN not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $FULL_PATH_TO_MTXHEADER ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_MTXHEADER not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $FULL_PATH_TO_MIDI ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $FULL_PATH_TO_MIDI not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $PATH_TO_MIDI/mscore.awk ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $PATH_TO_MIDI/mscore.awk not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $PATH_TO_MIDI/mscore.awk ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $PATH_TO_MIDI/mscore.awk not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ ! -f $PATH_TO_MIDI/hexxer ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
        echo ERROR: required software $PATH_TO_MIDI/hexxer not found
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
        exit 1;
	fi

	if [ -f $FULL_PATH_TO_SNDINFO ]; then
		echo
		echo "   *info*   sndinfo found at $FULL_PATH_TO_SNDINFO"
		echo
		using_sndinfo=1
	else
		echo
		echo "   *warning*   sndinfo not found at $FULL_PATH_TO_SNDINFO"
		echo
		using_sndinfo=0
	fi

#
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------


#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#	Script Variables
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#
#
#	-----------------
#	ASCORE score file.
#	-----------------
#
#	This is not ascore.awk, but the script that is passed to it
#
# 	command line usage:
#
#		ascore=<file.asc>
#
FILE_ASCORE=
#
#
#	------------
#	Score Node
#	------------
#
# 	command line usage:
#
#	node_score=MTX
#	node_score=BSCORE
#	node_score=CSOUND
#
CSND_NODE_SCORE=
#
#
#	------------
#	Mtx Node
#	------------
#
# 	command line usage:
#
#	node_mtx=MIDI
#
CSND_NODE_MTX=
#
#
#	------------
#	Midi Node
#	------------
#
# 	command line usage:
#
#	node_midi=RMIDI
#
CSND_NODE_MIDI=
#
#
#	------------
#	Rmidi Node
#	------------
#
# 	command line usage:
#
#	node_rmidi=RECORD
#
CSND_NODE_RMIDI=
#
#
#
#
#
#
#
#
#
#
#
#	------------
#	Bscore Node
#	------------
#
#
CSND_NODE_BSCORE=
#
#
#	------------------------------
#	Arguments to the record script
#	------------------------------
#
RECORD_NAME=out.wav
RECORD_DURATION=10.0
#
#
#
#
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#	Execution Flags
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#
#
run_ascore_flag=0
run_mtx_flag=0
run_midi_flag=0
run_rmidi_flag=0
run_record_flag=0
use_post_mtx="no"



#-------------------------------------------------------------------------
# Functions
#-------------------------------------------------------------------------



function Read_Arg()
{
	#
	# Evaluate 1 argument from the command-line
	#

	argstr=$1

	#
	#	does this arg have an "=" sign?
	#
	result=$(echo $argstr | grep =)
    len=${#result}

	if [ $len -ne 0 ]; then
		#
		# yes, there is an "=" sign
		#

		#
		# get the part of the string AFTER the equal sign
		#
		val=${argstr#*=}
        vlen=${#val}
        if [ $vlen -eq 0 ]; then
			echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
			echo
            echo ERROR: value in argument "("$argstr")" is zero-length;
			echo
			echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
            exit 1;
        fi

		#
		# get the part BEFORE the "=" sign
		#
		keywd=${argstr%=*}
        klen=${#keywd}
        if [ $klen -eq 0 ]; then
			echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
			echo
            echo ERROR: keyword in argument "("$argstr")" is zero-length
			echo
			echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
            exit 1;
        fi


		#
		#	set the variable
		#
		case $keywd in

			ascore )
				#
				# get the part of the string after the "=" sign
				#
				FILE_ASCORE=$val ;
				echo "   *info*   FILE_ASCORE = $FILE_ASCORE" ;
					;;

			csound )
				#
				# get the part of the string after the "=" sign
				#
				RUN_CSOUND=$val ;
				echo "   *info*   RUN_CSOUND = $RUN_CSOUND" ;
					;;

			node_score )
				#
				# get the part of the string after the "=" sign
				#
				CSND_NODE_SCORE=$val ;
				echo "   *info*   CSND_NODE_SCORE = $CSND_NODE_SCORE" ;
					;;

			node_mtx )
				#
				# get the part of the string after the "=" sign
				#
				CSND_NODE_MTX=$val ;
				echo "   *info*   CSND_NODE_MTX = $CSND_NODE_MTX" ;
					;;

			node_midi )
				#
				# get the part of the string after the "=" sign
				#
				CSND_NODE_MIDI=$val ;
				echo "   *info*   CSND_NODE_MIDI = $CSND_NODE_MIDI" ;
					;;

			node_rmidi )
				#
				# TBD TBD TBD TBD TBD TBD TBD TBD TBD TBD TBD TBD
				# get the part of the string after the "=" sign
				#
				#
				CSND_NODE_RMIDI=$val ;
				echo "   *info*   CSND_NODE_RMIDI = $CSND_NODE_RMIDI" ;
					;;

			record_name )
				#
				# get the part of the string after the "=" sign
				#
				RECORD_NAME=$val ;
				echo "   *info*   RECORD_NAME = $RECORD_NAME" ;
					;;

			record_duration )
				#
				# get the part of the string after the "=" sign
				#
				RECORD_DURATION=$val ;
				echo "   *info*   RECORD_DURATION = $RECORD_DURATION" ;
					;;

			post_mtx )
				#
				# get the part of the string after the "=" sign
				#
				use_post_mtx=$val ;
				echo "   *info*   RECORD_NAME = $RECORD_NAME" ;
					;;

			expect_rmidi )
				#
				# get the part of the string after the "=" sign
				#
				EXPECT_RMIDI=$val ;
				echo "   *info*   EXPECT_RMIDI = $EXPECT_RMIDI" ;
					;;

		esac


	fi # if an "=" sign




}
#------------------------ function Read_Arg -----------------------------



function Read_File()
{

	argstr=$1

	if [ -f "$argstr" ]; then

		while read file_arg file_comments; do

					Read_Arg $file_arg

		done < "$argstr"

	fi
}
#------------------------ function Read_File -----------------------------




function Cleanup()
{

	echo not impimented

}
#-------------------------- function Cleanup ----------------------------



#-------------------------------------------------------------------------
# End Functions
#-------------------------------------------------------------------------








#-------------------------------------------------------------------------
#	The Script Itself
#-------------------------------------------------------------------------
#
# Read the command line
# If the first argument is the name of a file, then that file has the list of
# node paths.  But command line arguments can still override it.
#
acnt=1
for arg in "$@"; do

	#
	#	evaluate the first argument
	#
	if [ $acnt -eq 1 ]; then
		#
		#	is it a file name?
		#
		if [ -f $arg ]; then
			#
			# there is such a file
			# read it
			#
			#
			Read_File $arg
			#
		else
			#
			# the first argument is a variable
			#
			Read_Arg $arg
			#
		fi
	else
		#
		# the 2nd and other arguments
		#
		Read_Arg $arg
		#
	fi


	let acnt=acnt+1

done



#
#	Variables should be set by the time we get here
#

#------------------------------------------------------------------------
#
# ASCORE
#
#------------------------------------------------------------------------
#
# check if there is really such a file
#
if [ -f "$FILE_ASCORE" ]; then
	#
	#
	#	do we have the ASCORE engine?
	#
	if [ -f "$FULL_PATH_TO_ASCORE" ]; then
		run_ascore_flag=1
	else
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: $FULL_PATH_TO_ASCORE is not found"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi

else

	echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
	echo
	echo "ERROR: ASCORE file ($FILE_ASCORE) not found"
	echo
	echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
	exit 1

fi
#
#
#
if [ $run_ascore_flag -ne 0 ]; then

	if [ -f hold.sco ]; then
		rm hold.sco
	fi

	echo
	echo "   *exec*   Running ASCORE on file $FILE_ASCORE"
	echo
	#
	# we run ascore.awk right here
	#
	cat $FILE_ASCORE | awk -f $FULL_PATH_TO_ASCORE  2> err.txt

	# check for errors
	if  ! awk -f $PATH_TO_CSNDSHELL/geterr.awk err.txt ; then
		echo "ERRORS in ASCORE -- see err.txt"
		exit 1
	else
		echo "SUCCESS running ASCORE"
	fi


	#
	#	get the name stem (part that comes before the "."
	#	
	main_file_stem=${FILE_ASCORE%.*}
	#
	#
	#
	#	done processing ASCORE
	#


	#
	#	run csound if the "csound=yes" argument given
	#
	if [ x$RUN_CSOUND == xyes ]; then

		echo "Running csound"
		if ! csound -d -o out.wav -W  ${main_file_stem}.orc  hold.sco  2> cerr.txt ; then
			echo "ERRORS in CSOUND -- see cerr.txt"
			exit 1
		fi

		# check for errors
		if  ! awk -f $PATH_TO_CSNDSHELL/geterr.awk cerr.txt ; then
			echo "ERRORS in CSOUND -- see cerr.txt"
			exit 1
		else
			echo
			echo "SUCCESS running CSOUND"
			echo
		fi

		if [ $using_sndinfo -eq 1 ]; then

			csound_duration=$( $FULL_PATH_TO_SNDINFO out.wav | awk '{ if ($8 == "seconds") {print $7} }' )
		else
			csound_duration="<duration>"
		fi

	fi


	#------------------------------------------------------------------------
	#
	# Node 1:  output from ASCORE
	#
	# What to do next with the output from ASCORE?
	#
	#
	case $CSND_NODE_SCORE in
	
		MTX )
			run_mtx_flag=1 ;
			echo "   *info*   NODE 1 (ascore) to MTX" ;
				;;
	
		BSCORE )
			run_bscore_flag=1 ;
			echo "   *info*   NODE 1 (ascore) to BSCORE" ;
				;;
	
	esac
	#------------------------------------------------------------------------

fi

#----------------------  END  ASCORE ------------------------------------
#------------------------------------------------------------------------





#------------------------------------------------------------------------
#
# MTX
#
# the script comes to this point when the "node_score=MTX" argument
# is used
#
#------------------------------------------------------------------------
#
if [ $run_mtx_flag -ne 0 ]; then

	#
	# check to see if there is really an output file from ascore
	#
	if [ ! -f "hold.sco" ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: MTX: no output file from ASCORE found"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi


	if [ ! -d "midi" ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: MTX: midi subdirectory does not exist---please create a subdirectory called midi in the working directory"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi

	#
	#   sort the Csound score (output from ascore)
	#
	FILE_SORTED_SCORE="${main_file_stem}_sorted.tmp"
	#
	# we call cs2sort.awk right here
	#
	echo
	echo "   *exec*   Running cs2sort.awk on ascore output"
	echo
	cat hold.sco | awk -f $FULL_PATH_TO_CS2SORT > $FILE_SORTED_SCORE
	#
	#	this is a temporary file that will be deleted
	#

	#
	#
	# get the arguments for cs2mtx.awk by parsing the output from ASCORE
	# with pre_cs2mtx.awk
	# these are taken from the CS2MTX statements in the ascore score
	#
	#
	echo
	echo "   *exec*   Running pre_cs2mtx.awk on ascore output"
	echo
	cat hold.sco | awk -f $FULL_PATH_TO_PRE_CS2MTX  > run_cs2mtx.sh

	#
	#	now run cs2mtx.awk by running the shell script run_cs2mtx.sh
	#	which was just created above
	#
	if [ -f run_cs2mtx.sh ]; then

		echo
		echo "   *exec*   Running run_cs2mtx.sh to create the MTX output"
		echo
		sh run_cs2mtx.sh $FILE_SORTED_SCORE $FULL_PATH_TO_CS2MTX $FULL_PATH_TO_MALIGN $FULL_PATH_TO_MTXHEADER $main_file_stem

	else
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: MTX: shell script called run_cs2mtx.sh was not created"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi


	#################################################################
	#	Conditionally USED
	#
	#
	#
	#	Post mtx script goes here
	#
	if [ ${use_post_mtx:=no} == "yes" ]; then
		echo
		echo "   *exec*   Running post_cs2mtx.awk on cs2mtx output"
		echo

		echo "#MTX file starts here" > eraseme.txt

		#
		# notice that 3 files are piped to AWK here
		#
		cat hold.sco eraseme.txt midi/${main_file_stem}.mtx | awk -f $FULL_PATH_TO_POST_CS2MTX  > tmp.mtx
		if [ -f tmp.mtx ]; then
			mv tmp.mtx midi/${main_file_stem}.mtx
		 fi

		#
		#	cleanup
		#
		if [ -f eraseme.txt ]; then
			rm eraseme.txt
		fi
	fi
	#################################################################


	zeroval=0
	if [ v$CSND_DEBUG == v$zeroval ]; then
		#
		#	clean up
		#
		if [ -f "$FILE_SORTED_SCORE" ]; then
			rm $FILE_SORTED_SCORE
		fi
		if [ -f "midi/body.mtx" ]; then
			rm midi/body.mtx
		fi
		if [ -f "midi/header.mtx" ]; then
			rm midi/header.mtx
		fi
		# rm run_cs2mtx.sh
	fi
	#
	#
	#
	#	done processing MTX
	#



	#------------------------------------------------------------------------
	#
	# Node 3:  output from MTX
	#
	# What to do next with the output from MTX?
	#
	#
	case $CSND_NODE_MTX in
	
		MIDI )
			run_midi_flag=1 ;
			echo "   *info*   NODE 3 (mtx) to MIDI" ;
				;;
	
	esac
	#------------------------------------------------------------------------
fi

#-------------------------  END  MTX ------------------------------------
#------------------------------------------------------------------------






#------------------------------------------------------------------------
#
# MIDI
#
# the script comes to this point when the "node_mtx=MIDI" argument
# is used
#
#------------------------------------------------------------------------
#
if [ $run_midi_flag -ne 0 ]; then

	#
	# check to see if there is really an output file from mtx
	#
	if [ ! -f "midi/${main_file_stem}.mtx" ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: MIDI: no output file from MTX found"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi


	if [ ! -f "$FULL_PATH_TO_MIDI" ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: MIDI: can't locate MIDI script $FULL_PATH_TO_MIDI"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi

	#
	# we call run2 right here
	#
	echo
	echo "   *exec*   Running run2 --mtx to midi converter-- on mtx output"
	echo
	$FULL_PATH_TO_MIDI $PATH_TO_MIDI "midi/${main_file_stem}.mtx" "midi/${main_file_stem}.mid"

	#
	# cleanup
	#
	if [ -f "tmp2.txt" ]; then
		rm tmp2.txt
	fi
	if [ -f "midi/tmp2.txt" ]; then
		rm midi/tmp2.txt
	fi
	if [ -f "tmp1.hex" ]; then
		rm tmp1.hex
	fi
	if [ -f "midi/tmp1.hex" ]; then
		rm midi/tmp1.hex
	fi
	if [ -f "tmp2.hex" ]; then
		rm tmp2.hex
	fi
	if [ -f "midi/tmp2.hex" ]; then
		rm midi/tmp2.hex
	fi

	#
	#
	#
	#	done processing MTX
	#


	#------------------------------------------------------------------------
	#
	# Node 4:  output from MIDI
	#
	# What to do next with the output from MIDI?
	#
	#
	case $CSND_NODE_MIDI in
	
		RMIDI )
			run_rmidi_flag=1 ;
			echo "   *info*   NODE 4 --midi-- to RMIDI" ;
				;;
	
	esac
	#------------------------------------------------------------------------

fi
#-------------------------  END  MIDI -----------------------------------
#------------------------------------------------------------------------



#------------------------------------------------------------------------
#
# RMIDI  (remote PC runs pmidi)
#
# the script comes to this point when the "node_midi=RMIDI" argument
# is used
#
#------------------------------------------------------------------------
#
if [ $run_rmidi_flag -ne 0 ]; then

	#
	# check to see if there is really an output file from midi
	#
	if [ ! -f "midi/${main_file_stem}.mid" ]; then
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@    ERROR     @~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		echo
		echo "ERROR: RMIDI: no output file from MIDI found"
		echo
		echo '@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@@~@~@~@'
		exit 1
	fi


	echo "   *info*   Trying to scp the .mid file to the remote pc....Enter Password:"
	echo


	#################################################################
	#	NOT USED
	#
	#	the pmidi command to be run on the remote machine
	#
	#	echo "pmidi -p 64:0 ${main_file_stem}.mid &" > midi/PMIDI.SH
	#	scp "midi/PMIDI.SH"  "$FULL_PATH_TO_RMIDI"
	#################################################################
	#
	#
	#
	# scp "midi/${main_file_stem}.mid"  "$FULL_PATH_TO_RMIDI"

	if [ -f $PATH_TO_CSNDSHELL/scp_expect.exp ]; then

        if [ x$EXPECT_RMIDI == xyes ]; then
            $PATH_TO_CSNDSHELL/scp_expect.exp "midi/${main_file_stem}.mid"  "$FULL_PATH_TO_RMIDI"
        else
            scp "midi/${main_file_stem}.mid"  "$FULL_PATH_TO_RMIDI"
        fi

    else

        scp "midi/${main_file_stem}.mid"  "$FULL_PATH_TO_RMIDI"

    fi

	#
	# the rest of this node is to run pmidi from an ssh session on the
	# remote pc, while at the same time starting a record session in
	# another window on this PC.
	#
	#	1.  Open an xterm, start an SSH session with the remote PC
	#	2.	Open an xterm in the mix subdirectory wher the recording
	#		will be done.
	#	3.	Start the recording on the local PC
	#	4.	Start the pmidi command on the remote PC
	#
	echo "-----------------------------------------------------------------"
	echo "	Now type"
	echo
	echo "	sh \$RECORD <output.wav> $csound_duration"
	echo
	echo "	on the local PC, and"
	echo
	echo "	PMIDI.SH ${main_file_stem}.mid"
	echo
	echo "	on the remote PC"
	echo "-----------------------------------------------------------------"
	echo


	#
	#	This is the END of the midi branch
	#

fi
#-------------------------  END  RMIDI ----------------------------------
#------------------------------------------------------------------------




