#
#   install.sh
#   Installer for csndshell
#
#   -------------------------------------------------------------------------
#   Revision history:
#   -------------------------------------------------------------------------
#   08-30-2009
#   A. Steffens Jr -- New file.
#   -------------------------------------------------------------------------
#
#   Requirements:
#
#   We expect to be running this file as the ROOT user from within
#   the path where the csndshell.sh file is located.
#

#
#   Find the path of the current directory
#
version_number=1.0.1
PATH_TO_CSNDSHELL=$(pwd)
echo PATH_TO_CSNDSHELL=$PATH_TO_CSNDSHELL

#
#   Get the current date string
#
date_string=$(date --rfc-3339='seconds')
fix_date=$(echo $date_string | awk \
          '{split($0,ary); str=sprintf("%s_%s",ary[1],ary[2]);\
           print str}' \
         )

#
#   Edit the global BASH profile to add the PATH_TO_CSNDSHELL
#   environment variable
#
if [ -f /etc/profile ]; then
    #
    #   Backup the shell init file
    #
    cp /etc/profile /etc/profile_$fix_date
    cp /etc/profile ./profile_$fix_date
    #
    #   Execute an AWK script for setting the PATH_TO_CSNDSHELL variable
    #
    cat /etc/profile | awk \
        'BEGIN { \
        has_variable=0; \
        has_csnd=0; \
        has_record=1; has_records=1; \
        has_mix=1; \
        } \
        $1 ~ /export/ { \
            if (match($2,/PATH_TO_CSNDSHELL/)==1) \
            { \
                print "export PATH_TO_CSNDSHELL="  path_to_csndshell ; \
                has_variable = 1 \
            } \
            else if (match($2,/RUN_CSND/)==1) \
            { \
                print "export RUN_CSND="  csnd ; \
                has_csnd = 1 \
            } \
            else if (match($2,/RECORD/)==1) \
            { \
                print "export RECORD="  record ; \
                has_record = 1 \
            } \
            else if (match($2,/RECORD_S/)==1) \
            { \
                print "export RECORD_S="  records ; \
                has_records = 1 \
            } \
            else if (match($2,/MIX/)==1) \
            { \
                print "export MIX="  mix ; \
                has_mix = 1 \
            } \
            else { print $0 } \
          next \
        } \
        { print $0 } \
        END { \
              if (has_variable==0) \
              { \
                print "export PATH_TO_CSNDSHELL="  path_to_csndshell ; \
              } \
              if (has_csnd==0) \
              { \
                print "export RUN_CSND="  csnd ; \
              } \
              if (has_record==0) \
              { \
                print "export RECORD="  record ; \
              } \
              if (has_records==0) \
              { \
                print "export RECORD_S="  records ; \
              } \
              if (has_mix==0) \
              { \
                print "export MIX="  mix ; \
              } \
        }' path_to_csndshell=$PATH_TO_CSNDSHELL \
        csnd=$PATH_TO_CSNDSHELL/csndshell.sh \
        record=$PATH_TO_CSNDSHELL/record/record.sh \
        records=$PATH_TO_CSNDSHELL/record/record_s.sh > \
        mix=$PATH_TO_CSNDSHELL/mixscore/mix.sh \
        /etc/tmp_profile
    #
    #   clean up
    #
    mv /etc/tmp_profile /etc/profile
else
    echo --------------------------------------------------------------------
    echo ERROR:
    echo Your system does not have /etc/profile, which means you may not be
    echo using the bash shell.  You will need to follow the Manual
    echo Installation instructions.  See the documentation for your logon
    echo shell for the locations of its initialization files.
    echo --------------------------------------------------------------------
    exit 1;
fi
echo Success updating /etc/profile

#
#   Set password in the scp_expect.exp script
#   set_my_password.sh is for each user to create
#
if [ -f ./set_my_password.sh ]; then
    sh ./set_my_password.sh
    echo Success restoring SSH password.
fi

#
#   Set execute permissions on scp_expect.exp
#
if [ -f ./scp_expect.exp ]; then
    chmod +x ./scp_expect.exp
else
    echo --------------------------------------------------------------------
    echo ERROR:
    echo scp_expect.exp file not found.
    echo --------------------------------------------------------------------
    exit 1;
fi


echo Installation complete.

